#ifndef EXTRACTCANDATA_H
#define EXTRACTCANDATA_H

#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <sstream>
#include <tuple>
#include <Eigen/Dense>

// Function to read timestamps from the file and save them as double
void readTimestamps(const std::string& filename, std::vector<double>& timestampsHigh, std::vector<double>& timestampsLow) {
    std::ifstream file(filename);
    std::string line;
    const std::string searchNsecs = "nsecs:";
    const std::string searchSecs = "secs:";

    if (!file.is_open()) {
        std::cerr << "Error opening file: " << filename << std::endl;
        return;
    }

    while (std::getline(file, line)) {
        std::size_t posNsecs = line.find(searchNsecs);
        if (posNsecs != std::string::npos) {
            std::istringstream iss0(line.substr(posNsecs + searchNsecs.length()));
            double timestampHigh;
            if (iss0 >> timestampHigh) {
                timestampsHigh.push_back(timestampHigh);
            }
        }

        std::size_t posSecs = line.find(searchSecs);
        if (posSecs != std::string::npos) {
            std::istringstream iss1(line.substr(posSecs + searchSecs.length()));
            double timestampLow;
            if (iss1 >> timestampLow) {
                timestampsLow.push_back(timestampLow);
            }
        }
    }

    file.close();
}

Eigen::MatrixXf extractAccelerationData(const std::string& filename) {
    std::ifstream file(filename);
    std::string line;
    Eigen::MatrixXf acc_matrix;

    if (file.is_open()) {
        std::vector<std::vector<float>> acc_data;
        while (getline(file, line)) {
            if (line.find("IMU_BIAS_CORRECTED_ACC_Signals:") != std::string::npos) {
                float acc_x, acc_y, acc_z;
                long long timestamp_s, timestamp_ns;
                getline(file, line);
                std::stringstream ss1(line.substr(line.find(": ") + 2));
                ss1 >> acc_x;
                getline(file, line);
                std::stringstream ss2(line.substr(line.find(": ") + 2));
                ss2 >> acc_y;
                getline(file, line);
                std::stringstream ss3(line.substr(line.find(": ") + 2));
                ss3 >> acc_z;
                getline(file, line);
                std::stringstream ss4(line.substr(line.find(": ") + 2));
                ss4 >> timestamp_s;
                getline(file, line);
                std::stringstream ss5(line.substr(line.find(": ") + 2));
                ss5 >> timestamp_ns;
                std::vector<float> row = {acc_x, acc_y, acc_z, static_cast<float>(timestamp_s), static_cast<float>(timestamp_ns)};
                acc_data.push_back(row);
            }
        }
        file.close();

        // Convert acc_data to Eigen::MatrixXf
        acc_matrix.resize(acc_data.size(), 5);
        for (size_t i = 0; i < acc_data.size(); ++i) {
            for (size_t j = 0; j < 5; ++j) {
                acc_matrix(i, j) = acc_data[i][j];
            }
        }
    } else {
        std::cerr << "Unable to open file" << std::endl;
    }

    return acc_matrix;
}


void extractSpeedAndYawRateData(const std::string& filename, Eigen::MatrixXf &speed_matrix, Eigen::MatrixXf &yaw_rate_matrix, Eigen::MatrixXf &ang_speed_matrix) {
    std::ifstream file(filename);
    std::string line;
    

    if (file.is_open()) {
        std::vector<std::vector<float>> speed_data;
        std::vector<std::vector<float>> ang_speed_data;
        std::vector<std::vector<float>> yaw_rate_data;
        while (getline(file, line)) {
            if (line.find("INS_SOL_RESULT_VEL_Signals:") != std::string::npos) {
                float speed, speed_x, speed_y, speed_z;                
                getline(file, line);
                std::stringstream ss1(line.substr(line.find(": ") + 2));
                ss1 >> speed;
                getline(file, line);
                std::stringstream ss2(line.substr(line.find(": ") + 2));
                ss2 >> speed_x;
                getline(file, line);
                std::stringstream ss3(line.substr(line.find(": ") + 2));
                ss3 >> speed_y;
                getline(file, line);
                std::stringstream ss4(line.substr(line.find(": ") + 2));
                ss4 >> speed_z;
                std::vector<float> rowOfSpeed = {speed, speed_x, speed_y, speed_z};
                speed_data.push_back(rowOfSpeed);
            }

            if (line.find("IMU_BIAS_CORRECTED_ANGRATE_Signals:") != std::string::npos) {
                float ang_speed_x, ang_speed_y, ang_speed_z;                
                getline(file, line);
                std::stringstream ss1(line.substr(line.find(": ") + 2));
                ss1 >> ang_speed_x;
                getline(file, line);
                std::stringstream ss2(line.substr(line.find(": ") + 2));
                ss2 >> ang_speed_y;
                getline(file, line);
                std::stringstream ss3(line.substr(line.find(": ") + 2));
                ss3 >> ang_speed_z;
                std::vector<float> rowOfAngSpeed = {ang_speed_x, ang_speed_y, ang_speed_z};
                ang_speed_data.push_back(rowOfAngSpeed);
            }

            if (line.find("IMU_BIAS_CORRECTED_ANGRATE_Signals:") != std::string::npos) {
                float yaw_rate;
                long long timestamp_s, timestamp_ns;
                getline(file, line);
                std::stringstream ss1(line.substr(line.find(": ") + 2));
                ss1 >> yaw_rate;
                std::vector<float> rowOfYawRate = {yaw_rate};
                yaw_rate_data.push_back(rowOfYawRate);
            }
        }
        file.close();

        // Convert speed_data to Eigen::MatrixXf
        speed_matrix.resize(speed_data.size(), 4);
        for (size_t i = 0; i < speed_data.size(); ++i) {
            for (size_t j = 0; j < 4; ++j) {
                speed_matrix(i, j) = speed_data[i][j];
            }
        }
        // Convert speed_data to Eigen::MatrixXf
        ang_speed_matrix.resize(ang_speed_data.size(), 3);
        for (size_t i = 0; i < ang_speed_data.size(); ++i) {
            for (size_t j = 0; j < 3; ++j) {
                ang_speed_matrix(i, j) = ang_speed_data[i][j];
            }
        }
        // Convert yaw_rate_data to Eigen::MatrixXf
        yaw_rate_matrix.resize(yaw_rate_data.size(), 1);
        for (size_t i = 0; i < yaw_rate_data.size(); ++i) {
            for (size_t j = 0; j < 1; ++j) {
                yaw_rate_matrix(i, j) = yaw_rate_data[i][j];
            }
        }
    } else {
        std::cerr << "Unable to open file" << std::endl;
    }

    return;
}


Eigen::MatrixXf extractOrientationData(const std::string& filename) {
    std::ifstream file(filename);
    std::string line;
    Eigen::MatrixXf orientation_matrix;

    if (file.is_open()) {
        std::vector<std::vector<float>> orientation_data;
        while (getline(file, line)) {
            if (line.find("INS_SOL_RESULT_ANGLE_Signals:") != std::string::npos) {
                float orientation, orientation_x, orientation_y, orientation_z;
                long long timestamp_s, timestamp_ns;
                getline(file, line);
                std::stringstream ss1(line.substr(line.find(": ") + 2));
                ss1 >> orientation;
                getline(file, line);
                std::stringstream ss2(line.substr(line.find(": ") + 2));
                ss2 >> orientation_z;
                getline(file, line);
                std::stringstream ss3(line.substr(line.find(": ") + 2));
                ss3 >> orientation_y;
                getline(file, line);
                std::stringstream ss4(line.substr(line.find(": ") + 2));
                ss4 >> orientation_x;
                std::vector<float> row = {orientation_z, orientation_y, orientation_x};
                orientation_data.push_back(row);
            }
        }
        file.close();

        // Convert acc_data to Eigen::MatrixXf
        orientation_matrix.resize(orientation_data.size(), 3);
        for (size_t i = 0; i < orientation_data.size(); ++i) {
            for (size_t j = 0; j < 3; ++j) {
                orientation_matrix(i, j) = orientation_data[i][j];
            }
        }
    } else {
        std::cerr << "Unable to open file" << std::endl;
    }

    return orientation_matrix;
}


void extractPositionData(const std::string& filename,
                                Eigen::MatrixXf &position_z_matrix,
                                Eigen::MatrixXf &position_y_matrix,
                                Eigen::MatrixXf &position_x_matrix) {
    std::ifstream file(filename);
    std::string line;
    

    if (file.is_open()) {
        std::vector<std::vector<float>> position_z_data;
        std::vector<std::vector<float>> position_y_data;
        std::vector<std::vector<float>> position_x_data;
        while (getline(file, line)) {
            if (line.find("INS_SOL_RESULT_ALTITUDE_Signals:") != std::string::npos) {
                float position_z;                
                getline(file, line);
                std::stringstream ss1(line.substr(line.find(": ") + 2));
                ss1 >> position_z;
                std::vector<float> rowOfPositionZ = {position_z};
                position_z_data.push_back(rowOfPositionZ);
            }

            if (line.find("INS_SOL_RESULT_LATITUDE_Signals:") != std::string::npos) {
                float position_y;
                getline(file, line);
                std::stringstream ss1(line.substr(line.find(": ") + 2));
                ss1 >> position_y;
                std::vector<float> rowOfPositionY = {position_y};
                position_y_data.push_back(rowOfPositionY);
            }

            if (line.find("INS_SOL_RESULT_LONGITUDE_Signals:") != std::string::npos) {
                float position_x;
                getline(file, line);
                std::stringstream ss1(line.substr(line.find(": ") + 2));
                ss1 >> position_x;
                std::vector<float> rowOfPositionX = {position_x};
                position_x_data.push_back(rowOfPositionX);
            }
        }
        file.close();

        // Convert position_z_data to Eigen::MatrixXf
        position_z_matrix.resize(position_z_data.size(), 1);
        for (size_t i = 0; i < position_z_data.size(); ++i) {
            for (size_t j = 0; j < 1; ++j) {
                position_z_matrix(i, j) = position_z_data[i][j];
            }
        }
        // Convert position_y_data to Eigen::MatrixXf
        position_y_matrix.resize(position_y_data.size(), 1);
        for (size_t i = 0; i < position_y_data.size(); ++i) {
            for (size_t j = 0; j < 1; ++j) {
                position_y_matrix(i, j) = position_y_data[i][j];
            }
        }
        // Convert position_x_data to Eigen::MatrixXf
        position_x_matrix.resize(position_x_data.size(), 1);
        for (size_t i = 0; i < position_x_data.size(); ++i) {
            for (size_t j = 0; j < 1; ++j) {
                position_x_matrix(i, j) = position_x_data[i][j];
            }
        }
    } else {
        std::cerr << "Unable to open file" << std::endl;
    }

    return;
}


std::vector<std::tuple<float, float, float, long long, long long>> extractAccData(const std::string& filename) {
    std::ifstream file(filename);
    std::string line;
    std::vector<std::tuple<float, float, float, long long, long long>> acc_data;

    if (file.is_open()) {
        while (getline(file, line)) {
            if (line.find("IMU_BIAS_CORRECTED_ACC_Signals:") != std::string::npos) {
                float acc_x, acc_y, acc_z;
                long long timestamp_s, timestamp_ns;
                getline(file, line);
                std::stringstream ss1(line.substr(line.find(": ") + 2));
                ss1 >> acc_x;
                getline(file, line);
                std::stringstream ss2(line.substr(line.find(": ") + 2));
                ss2 >> acc_y;
                getline(file, line);
                std::stringstream ss3(line.substr(line.find(": ") + 2));
                ss3 >> acc_z;
                getline(file, line);
                std::stringstream ss4(line.substr(line.find(": ") + 2));
                ss4 >> timestamp_s;
                getline(file, line);
                std::stringstream ss5(line.substr(line.find(": ") + 2));
                ss5 >> timestamp_ns;
                acc_data.emplace_back(acc_x, acc_y, acc_z, timestamp_s, timestamp_ns);
            }
        }
        file.close();
    } else {
        std::cerr << "Unable to open file" << std::endl;
    }

    return acc_data;
}

#endif
